import pytest
from ewoksorange.bindings import ows_to_ewoks
from ewoksorange.orange_version import ORANGE_VERSION
from ewokscore import execute_graph

try:
    from importlib import resources
except ImportError:
    import importlib_resources as resources


def test_sumtask_tutorial1_with_qt(ewoks_orange_canvas):
    from orangecontrib.ewoks_example_category import tutorials

    with resources.path(tutorials, "sumtask_tutorial1.ows") as filename:
        assert_sumtask_tutorial_with_qt(ewoks_orange_canvas, filename)


@pytest.mark.skipif(
    ORANGE_VERSION != ORANGE_VERSION.latest_orange, reason="Requires Orange3 widgets"
)
def test_sumtask_tutorial1_without_qt(register_ewoks_example_addons):
    from orangecontrib.ewoks_example_category import tutorials

    with resources.path(tutorials, "sumtask_tutorial1.ows") as filename:
        assert_sumtask_tutorial_without_qt(filename)


def test_sumtask_tutorial2_with_qt(ewoks_orange_canvas):
    from orangecontrib.evaluate.ewoks_example_submodule import tutorials

    with resources.path(tutorials, "sumtask_tutorial2.ows") as filename:
        assert_sumtask_tutorial_with_qt(ewoks_orange_canvas, filename)


def test_sumtask_tutorial2_without_qt(ewoks_orange_canvas):
    from orangecontrib.evaluate.ewoks_example_submodule import tutorials

    with resources.path(tutorials, "sumtask_tutorial2.ows") as filename:
        assert_sumtask_tutorial_without_qt(filename)


def test_sumtask_tutorial3_with_qt(ewoks_orange_canvas):
    from orangecontrib.ewoks_example_supercategory.ewoks_example_subcategory import (
        tutorials,
    )

    with resources.path(tutorials, "sumtask_tutorial3.ows") as filename:
        assert_sumtask_tutorial_with_qt(ewoks_orange_canvas, filename)


def test_sumtask_tutorial3_without_qt(ewoks_orange_canvas):
    from orangecontrib.ewoks_example_supercategory.ewoks_example_subcategory import (
        tutorials,
    )

    with resources.path(tutorials, "sumtask_tutorial3.ows") as filename:
        assert_sumtask_tutorial_without_qt(filename)


def test_list_operations_with_qt(ewoks_orange_canvas):
    from orangecontrib.list_operations import tutorials

    with resources.path(tutorials, "sumlist_tutorial.ows") as filename:
        assert_sumlist_tutorial_with_qt(ewoks_orange_canvas, filename)


def test_list_operations_without_qt(ewoks_orange_canvas):
    from orangecontrib.list_operations import tutorials

    with resources.path(tutorials, "sumlist_tutorial.ows") as filename:
        assert_sumlist_tutorial_without_qt(filename)


@pytest.mark.skipif(
    ORANGE_VERSION != ORANGE_VERSION.latest_orange, reason="Requires Orange3 widgets"
)
def test_mixed_tutorial1_with_qt(ewoks_orange_canvas):
    from orangecontrib.ewoks_example_category import tutorials

    with resources.path(tutorials, "mixed_tutorial1.ows") as filename:
        assert_mixed_tutorial_with_qt(ewoks_orange_canvas, filename)


@pytest.mark.skipif(
    ORANGE_VERSION != ORANGE_VERSION.latest_orange, reason="Requires Orange3 widgets"
)
def test_mixed_tutorial1_without_qt(register_ewoks_example_addons):
    from orangecontrib.ewoks_example_category import tutorials

    with resources.path(tutorials, "mixed_tutorial1.ows") as filename:
        assert_mixed_tutorial_without_qt(filename)


def assert_sumtask_tutorial_with_qt(ewoks_orange_canvas, filename):
    """Execute workflow using the Qt widgets and signals"""
    ewoks_orange_canvas.load_ows(str(filename))
    ewoks_orange_canvas.wait_widgets()
    widgets = list(ewoks_orange_canvas.widgets_from_name("task6"))
    results = widgets[0].task_output_values
    assert results == {"result": 16}


def assert_sumtask_tutorial_without_qt(filename):
    """Execute workflow after converting it to an ewoks workflow"""
    graph = ows_to_ewoks(filename)
    results = execute_graph(graph)
    assert results["5"].output_values == {"result": 16}


def assert_sumlist_tutorial_with_qt(ewoks_orange_canvas, filename):
    """Execute workflow using the Qt widgets and signals"""
    ewoks_orange_canvas.load_ows(str(filename))

    # Remove artificial delay for this test
    for widget in ewoks_orange_canvas.iter_widgets():
        if "delay" in widget.default_inputs:
            widget.default_inputs["delay"] = 0

    wgenerator = list(ewoks_orange_canvas.widgets_from_name("List generator"))[0]
    wgenerator.defaultInputsHaveChanged()
    results = wgenerator.task_output_values
    listsum = sum(results["list"])

    ewoks_orange_canvas.wait_widgets()

    widgets = list(ewoks_orange_canvas.widgets_from_name("Print list sum"))
    widgets = list(ewoks_orange_canvas.widgets_from_name("Print list sum (1)"))
    widgets = list(ewoks_orange_canvas.widgets_from_name("Print list sum (2)"))
    for w in widgets:
        results = {name: var.value for name, var in w.task_inputs.items()}
        assert results == {"sum": listsum}


def assert_sumlist_tutorial_without_qt(filename):
    """Execute workflow after converting it to an ewoks workflow"""
    graph = ows_to_ewoks(filename)

    # Remove artificial delay for this test
    for attrs in graph.graph.nodes.values():
        for adict in attrs.get("default_inputs", list()):
            if adict["name"] == "delay":
                adict["value"] = 0

    results = execute_graph(graph, results_of_all_nodes=True)
    listsum = sum(results["0"].output_values["list"])
    for i in [4, 5, 6]:
        assert results[str(i)].input_values == {"sum": listsum}


def assert_mixed_tutorial_with_qt(ewoks_orange_canvas, filename):
    """Execute workflow using the Qt widgets and signals"""
    ewoks_orange_canvas.load_ows(str(filename))
    ewoks_orange_canvas.wait_widgets()
    widgets = list(ewoks_orange_canvas.widgets_from_name("Adder2"))
    results = widgets[0].task_output_values
    assert results == {"result": 3}


def assert_mixed_tutorial_without_qt(filename):
    """Execute workflow after converting it to an ewoks workflow"""
    graph = ows_to_ewoks(filename)
    results = execute_graph(graph)
    assert results["1"].output_values == {"result": 3}



import gc
import logging
import pytest
from ewoksorange.registration import register_addon_package
from ewoksorange.bindings.qtapp import qtapp_context
from ewoksorange.bindings.qtapp import get_all_qtwidgets
from ewoksorange.canvas.handler import OrangeCanvasHandler
from .examples import ewoks_example_1_addon
from .examples import ewoks_example_2_addon
from ewoksorange import ewoks_addon
from ewoksorange.orange_version import ORANGE_VERSION


logger = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def register_ewoks_example_1_addon():
    register_addon_package(ewoks_example_1_addon)
    yield


@pytest.fixture(scope="session")
def register_ewoks_example_2_addon():
    register_addon_package(ewoks_example_2_addon)
    yield


@pytest.fixture(scope="session")
def register_ewoks_example_addons(
    register_ewoks_example_1_addon, register_ewoks_example_2_addon
):
    yield


def global_cleanup_orange():
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        pass
    else:
        from orangecanvas.document.suggestions import Suggestions

        Suggestions.instance = None


def global_cleanup_pytest():
    for obj in gc.get_objects():
        if isinstance(obj, logging.LogRecord):
            obj.exc_info = None  # traceback keeps frames which keep locals


def collect_garbage(app):
    app.processEvents()
    while gc.collect():
        app.processEvents()


@pytest.fixture(scope="session")
def qtapp():
    with qtapp_context() as app:
        yield app
    collect_garbage(app)
    global_cleanup_orange()
    global_cleanup_pytest()
    collect_garbage(app)
    warn_qtwidgets_alive()


@pytest.fixture(scope="session")
def ewoks_orange_canvas(qtapp, register_ewoks_example_addons):
    register_addon_package(ewoks_addon)
    with OrangeCanvasHandler() as handler:
        yield handler


def warn_qtwidgets_alive():
    widgets = get_all_qtwidgets()
    if widgets:
        logger.warning(
            "%d remaining widgets after tests:\n %s",
            len(widgets),
            "\n ".join(map(str, widgets)),
        )



from ewoksorange.registration import get_owwidget_descriptions


def test_orangecontrib_packages(register_ewoks_example_addons):
    from orangecontrib import ewoks_example_supercategory  # noqa F401
    from orangecontrib.evaluate import ewoks_example_submodule  # noqa F401
    from orangecontrib import ewoks_example_category  # noqa F401
    from orangecontrib import list_operations  # noqa F401


def test_discover_widgets(register_ewoks_example_addons):
    widgets = get_owwidget_descriptions()
    discovered = set(w.qualified_name for w in widgets)
    expected = {
        "orangecontrib.ewoks_example_supercategory.ewoks_example_subcategory.adder1.Adder1",
        "orangecontrib.ewoks_example_supercategory.ewoks_example_subcategory.adder2.Adder2",
        "orangecontrib.evaluate.ewoks_example_submodule.adder1.Adder1",
        "orangecontrib.evaluate.ewoks_example_submodule.adder2.Adder2",
        "orangecontrib.ewoks_example_category.adder1.Adder1",
        "orangecontrib.ewoks_example_category.adder2.Adder2",
        "orangecontrib.list_operations.listgenerator.ListGenerator",
        "orangecontrib.list_operations.print_sum.PrintSumOW",
        "orangecontrib.list_operations.sumlist_one_thread.SumListOneThread",
        "orangecontrib.list_operations.sumlist_several_thread.SumListSeveralThread",
        "orangecontrib.list_operations.sumlist_stack.SumListWithTaskStack",
    }
    assert (discovered & expected) == expected



from typing import List, Tuple

from ewokscore import Task
from ewokscore.utils import qualname
from ewokscore.utils import import_qualname
from ewokscore.variable import value_from_transfer

from .qtapp import ensure_qtapp
from .qtapp import process_qtapp_events
from .qtapp import QtEvent
from .owwidgets import is_ewoks_widget_class
from .owwidgets import is_native_widget_class
from . import owsignals
from . import owsettings
from .owsignal_manager import SignalManagerWithoutScheme
from .owsignal_manager import set_input_value
from ..orange_version import ORANGE_VERSION

__all__ = ["OWWIDGET_TASKS_GENERATOR"]


def owwidget_task_wrapper(widget_qualname: str) -> Task:
    """Create a task that does the computation through an orange widget.
    When the widget is an ewoks widget, still use the widget and not
    the corresponding task class directly.
    """
    widget_class = import_qualname(widget_qualname)
    registry_name = widget_qualname + ".wrapper"
    if registry_name in Task.get_subclass_names():
        return Task.get_subclass(registry_name)

    if is_ewoks_widget_class(widget_class):
        return _ewoks_owwidget_task_wrapper(registry_name, widget_class)
    elif is_native_widget_class(widget_class):
        return _native_owwidget_task_wrapper(registry_name, widget_class)
    else:
        raise TypeError(widget_class, "expected to be an OWWidget")


OWWIDGET_TASKS_GENERATOR = qualname(owwidget_task_wrapper)


def _ewoks_owwidget_task_wrapper(registry_name, widget_class) -> Task:
    """Wrap an Ewoks widget with an Ewoks task"""
    all_input_names = widget_class.input_names()
    try:
        ewokstaskclass = widget_class.ewokstaskclass
        input_names = ewokstaskclass.required_input_names()
        optional_input_names = ewokstaskclass.optional_input_names()
        expected = set(input_names) | set(optional_input_names)
        assert all_input_names == expected
    except AttributeError:
        input_names = all_input_names
        optional_input_names = tuple()
    output_names = widget_class.output_names()

    class WrapperTask(
        Task,
        input_names=input_names,
        optional_input_names=optional_input_names,
        output_names=output_names,
        registry_name=registry_name,
    ):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.outputsReceived = None
            self.widget = None

        def run(self):
            ensure_qtapp()
            self.outputsReceived = QtEvent()
            self.widget = widget_class()
            try:
                self.widget.task_output_changed_callbacks.add(self.receiveOutputSend)

                # Call the input setters
                values = self.input_values

                if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
                    signals = self.widget.inputs
                else:
                    signals = self.widget.get_signals("inputs")
                orange_to_ewoks = owsignals.get_orange_to_ewoks_mapping(
                    widget_class, "inputs"
                )

                for index, signal in enumerate(signals):
                    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
                        key = orange_to_ewoks.get(signal.name, signal.name)
                    else:
                        key = signal.ewoksname
                    set_input_value(self.widget, signal, values[key], index)

                # Start calculation
                self.widget.handleNewSignals()

                # Wait for the result
                self.outputsReceived.wait()
            finally:
                self.widget.close()
                self.widget = None
                self.outputsReceived = None

        def receiveOutputSend(self):
            try:
                for name, value in self.widget.task_output_values.items():
                    self.output_variables[name].value = value
            finally:
                if self.outputsReceived is not None:
                    self.outputsReceived.set()

    return WrapperTask


def _native_owwidget_task_wrapper(registry_name, widget_class) -> Task:
    """Wrap a native Orange widget with an Ewoks task"""
    input_signals = owsignals.get_signals(widget_class.Inputs)
    optional_input_names = set(input_signals.keys())
    output_signals = owsignals.get_signals(widget_class.Outputs)
    output_names = set(output_signals.keys())
    orange_to_ewoks_namemap = {
        ewoks_name: signal.name for ewoks_name, signal in output_signals.items()
    }
    input_names = tuple()
    setting_names = list(owsettings.get_settings(widget_class))

    class WrapperTask(
        Task,
        input_names=input_names,
        optional_input_names=optional_input_names,
        output_names=output_names,
        registry_name=registry_name,
    ):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            self.outputsReceived = None

        def run(self):
            input_list, settings_dict = self._parse_input_values()

            # Create widget with the proper settings
            ensure_qtapp()
            widget = instantiate_owwidget(
                widget_class,
                signal_manager=SignalManagerWithoutScheme(),
                stored_settings=settings_dict,
            )
            try:
                # Call input setters
                for index, (signal, value) in enumerate(input_list):
                    set_input_value(widget, signal, value, index)

                # Start calculation
                widget.handleNewSignals()

                # Wait for the result
                process_qtapp_events()

                for ewoks_name, orange_name in orange_to_ewoks_namemap.items():
                    value = widget.signalManager.get_output_value(
                        widget, orange_name, timeout=None
                    )
                    self.output_variables[ewoks_name].value = value
            finally:
                widget.close()

        def _parse_input_values(self) -> Tuple[List, dict]:
            values = self.input_values

            used_values = set()
            settings_dict = dict()
            input_list = list()

            # Values corresponding to settings
            for ewoksname in setting_names:
                if ewoksname not in values:
                    continue
                used_values.add(ewoksname)
                value = value_from_transfer(values[ewoksname])
                if value is not self.MISSING_DATA:
                    settings_dict[ewoksname] = value

            # Values corresponding to inputs
            for signal in widget_class.get_signals("inputs"):
                ewoksname = owsignals.signal_orange_to_ewoks_name(
                    widget_class, "inputs", signal.name
                )
                if ewoksname not in values:
                    continue
                used_values.add(ewoksname)
                value = value_from_transfer(values[ewoksname])
                if value is self.MISSING_DATA:
                    value = None
                input_list.append((signal, value))

            # Node properties not corresponding to settings or inputs
            # are used in settings migration
            unused_values = set(values.keys()) - used_values
            for ewoksname in unused_values:
                value = value_from_transfer(values[ewoksname])
                if value is not self.MISSING_DATA:
                    settings_dict[ewoksname] = value

            return input_list, settings_dict

    return WrapperTask


def instantiate_owwidget(widget_class, signal_manager=None, stored_settings=None):
    widget = widget_class.__new__(
        widget_class, signal_manager=signal_manager, stored_settings=stored_settings
    )
    widget.__init__()
    return widget



import gc
import time
import signal
from typing import Optional
from contextlib import contextmanager
from AnyQt.QtWidgets import QApplication
from AnyQt.QtCore import QEventLoop

APP = None


def ensure_qtapp() -> Optional[QApplication]:
    """Create a Qt application without event loop when no application is running"""
    global APP
    if APP is not None:
        return

    # GUI application
    APP = QApplication.instance()
    if APP is not None:
        return APP

    # Allow termination with CTRL + C
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    # Application without event loop (APP.exec() is not called)
    APP = QApplication([])
    return APP


def close_qtapp():
    """Close the Qt application created by ensure_qtapp"""
    global APP
    if APP is None:
        return
    APP.processEvents()
    while gc.collect():
        APP.processEvents()
    APP.exit()
    APP = None


@contextmanager
def qtapp_context():
    qtapp = ensure_qtapp()
    try:
        yield qtapp
    finally:
        close_qtapp()


def get_qtapp() -> Optional[QApplication]:
    return QApplication.instance()


def process_qtapp_events():
    """Process all pending Qt events when a Qt event loop is running"""
    global APP
    if APP is None:
        return
    while APP.hasPendingEvents():
        APP.processEvents()


class QtEvent:
    """Event that also works for Qt applications with an event loop
    that need to run manually"""

    def __init__(self):
        self.__flag = False

    def wait(self, timeout=None):
        """Processes events associated to the calling thread while waiting"""
        global APP
        if timeout is not None:
            t0 = time.time()
        while not self.__flag:
            if APP is None:
                time.sleep(0.1)
            else:
                APP.processEvents(QEventLoop.AllEvents, 100)
            if timeout is not None:
                secs = time.time() - t0
                if secs <= 0:
                    return False
        return True

    def set(self):
        self.__flag = True

    def clear(self):
        self.__flag = False


def get_all_qtwidgets():
    app = get_qtapp()
    if app is None:
        return list()

    sapp = str(type(app))
    if "PyQt5" in sapp:
        from PyQt5.sip import ispycreated as createdByPython  # noqa
    elif "PySide2" in sapp:
        from PySide2.shiboken2 import createdByPython  # noqa
    else:
        raise RuntimeError(f"'{sapp}' not supported")

    return [widget for widget in app.allWidgets() if createdByPython(widget)]



"""Each Orange3 Addon install entry-points for widgets and tutorials.

Widget discovery is done in `orangecanvas.registry.discovery.WidgetDiscovery`
"""

import pkgutil
import importlib
from typing import List, Optional
import pkg_resources
import logging

from .orange_version import ORANGE_VERSION

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    from orangewidget.canvas.discovery import WidgetDiscovery
    from orangecanvas.registry.base import WidgetRegistry
    from orangecanvas.registry.description import WidgetDescription
    from orangecanvas.registry.description import InputSignal
    from orangecanvas.registry.description import OutputSignal
    from orangecanvas.registry import global_registry
    from orangecanvas.registry.utils import category_from_package_globals

    def get_widget_description(widget_class) -> WidgetDescription:
        widget_cls_name = widget_class.__name__

        qualified_name = "%s.%s" % (widget_class.__module__, widget_cls_name)

        inputs = [
            InputSignal(s.name, s.type, s.handler, s.flags, s.id, s.doc)
            for s in widget_class.inputs
        ]
        outputs = [
            OutputSignal(s.name, s.type, s.flags, s.id, s.doc)
            for s in widget_class.outputs
        ]
        # Convert all signal types into qualified names.
        # This is to prevent any possible import problems when cached
        # descriptions are unpickled (the relevant code using this lists
        # should be able to handle missing types better).
        for s in inputs + outputs:
            if isinstance(s.type, type):
                s.type = "%s.%s" % (s.type.__module__, s.type.__name__)

        return WidgetDescription(
            name=widget_class.name,
            id=widget_class.id,
            version=widget_class.version,
            description=widget_class.description,
            long_description=widget_class.long_description,
            qualified_name=qualified_name,
            inputs=inputs,
            outputs=outputs,
            author=widget_class.author,
            author_email=widget_class.author_email,
            maintainer=widget_class.maintainer,
            maintainer_email=widget_class.maintainer_email,
            help=widget_class.help,
            help_ref=widget_class.help_ref,
            url=widget_class.url,
            keywords=widget_class.keywords,
            priority=widget_class.priority,
            icon=widget_class.icon,
            background=widget_class.background,
            replaces=widget_class.replaces,
        )

    NATIVE_WIDGETS_PROJECT = "oasys1"
else:
    from orangewidget.workflow.discovery import WidgetDiscovery
    from orangecanvas.registry.base import WidgetRegistry
    from orangecanvas.registry import WidgetDescription
    from orangecanvas.registry import global_registry
    from orangecanvas.registry.utils import category_from_package_globals

    if ORANGE_VERSION == ORANGE_VERSION.latest_orange:
        NATIVE_WIDGETS_PROJECT = "orange3"
    else:
        NATIVE_WIDGETS_PROJECT = "orange"

from ewoksorange import setuptools
from .canvas.utils import get_orange_canvas

logger = logging.getLogger(__name__)


def get_distribution(distroname):
    try:
        return pkg_resources.get_distribution(distroname)
    except Exception:
        return None


def add_entry_points(distribution, entry_points):
    """Add entry points to a package distribution

    :param dict entry_points: mapping of "groupname" to a list of entry points
                              ["ep1 = destination1", "ep1 = destination2", ...]
    """
    if isinstance(distribution, str):
        distroname = distribution
        dist = get_distribution(distroname)
        if dist is None:
            logger.error(
                "Distribution '%s' not found. Existing distributions:\n %s",
                distroname,
                list(pkg_resources.working_set.by_key.keys()),
            )
            raise pkg_resources.DistributionNotFound(distroname, [repr("ewoksorange")])
    else:
        dist = distribution
        distroname = dist.project_name

    entry_map = dist.get_entry_map()
    for group, lst in entry_points.items():
        group_map = entry_map.setdefault(group, dict())
        for entry_point in lst:
            ep = pkg_resources.EntryPoint.parse(entry_point, dist=dist)
            if ep.name in group_map:
                raise ValueError(
                    f"Entry point {repr(ep.name)} already exists in group {repr(group)} of distribution {repr(distroname)}"
                )
            group_map[ep.name] = ep
            logger.debug("Dynamically add entry point for '%s': %s", distroname, ep)


def create_fake_distribution(distroname, location):
    distroname = pkg_resources.safe_name(distroname)
    dist = get_distribution(distroname)
    if dist is not None:
        raise RuntimeError(
            f"A distribution with the name {repr(distroname)} already exists"
        )
    if isinstance(location, list):
        location = location[0]
    from ewoksorange import __version__

    dist = pkg_resources.Distribution(
        location=location, project_name=distroname, version=__version__
    )
    pkg_resources.working_set.add(dist)
    return dist


def get_subpackages(package):
    for pkginfo in pkgutil.walk_packages(package.__path__, package.__name__ + "."):
        if pkginfo.ispkg:
            yield importlib.import_module(pkginfo.name)


def register_addon_package(package, distroname: Optional[str] = None):
    """An Orange Addon package which has not been installed."""
    entry_points = dict()
    packages = list(get_subpackages(package))
    if not distroname:
        distroname = package.__name__.split(".")[-1]
    setuptools.update_entry_points(packages, entry_points, distroname)
    dist = create_fake_distribution(distroname, package.__path__)
    add_entry_points(dist, entry_points)


def widget_discovery(discovery, distroname, subpackages):
    dist = pkg_resources.get_distribution(distroname)
    for pkg in subpackages:
        discovery.process_category_package(pkg, distribution=dist)


def iter_entry_points(group):
    """Do not include native orange entry points"""
    for ep in pkg_resources.iter_entry_points(group):
        if ep.dist.project_name.lower() != NATIVE_WIDGETS_PROJECT:
            yield ep


def global_registry_objects() -> List[WidgetRegistry]:
    registry_objects = list()
    scene = None
    canvas = get_orange_canvas()
    if canvas is not None:
        scene = canvas.current_document()
        reg = canvas.widget_registry
        if reg is not None:
            registry_objects.append(reg)
    if ORANGE_VERSION != ORANGE_VERSION.oasys_fork and scene is not None:
        reg = scene.registry()
        if reg is not None:
            registry_objects.append(reg)
    if not registry_objects:
        reg = global_registry()
        if reg is not None:
            registry_objects.append(reg)
    return registry_objects


def global_discovery_objects() -> List[WidgetDiscovery]:
    return [WidgetDiscovery(reg) for reg in global_registry_objects()]


def local_discovery_object() -> WidgetDiscovery:
    return WidgetDiscovery(WidgetRegistry())


def get_owwidget_descriptions():
    """Do not include native orange widgets"""
    disc = local_discovery_object()
    disc.run(iter_entry_points(setuptools.WIDGET_GROUP))
    return disc.registry.widgets()


def get_owwidget_description(
    widget_class, package_name: str, category_name: str, project_name: str
):
    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        description = get_widget_description(widget_class)
    else:
        kwargs = widget_class.get_widget_description()
        description = WidgetDescription(**kwargs)
    description.package = setuptools.orangecontrib_qualname(package_name)
    description.category = widget_class.category or category_name
    description.project_name = project_name
    return description


def get_owcategory_description(
    package_name: str, category_name: str, project_name: str
):
    description = category_from_package_globals(package_name)
    description.name = category_name
    description.project_name = project_name
    return description


def register_owcategory(
    package_name: str,
    category_name: str,
    project_name: str,
    discovery_object: Optional[WidgetDiscovery] = None,
):
    description = get_owcategory_description(package_name, category_name, project_name)
    if discovery_object is None:
        for discovery_object in global_discovery_objects():
            discovery_object.handle_category(description)
    else:
        discovery_object.handle_category(description)


def register_owwidget(
    widget_class,
    package_name: str,
    category_name: str,
    project_name: str,
    discovery_object: Optional[WidgetDiscovery] = None,
):
    register_owcategory(
        package_name, category_name, project_name, discovery_object=discovery_object
    )
    description = get_owwidget_description(
        widget_class, package_name, category_name, project_name
    )

    logger.debug("Register widget: %s", description.qualified_name)
    if discovery_object is None:
        for discovery_object in global_discovery_objects():
            if (
                discovery_object.registry is not None
                and discovery_object.registry.has_widget(description.qualified_name)
            ):
                continue
            discovery_object.handle_widget(description)
    else:
        if (
            discovery_object.registry is not None
            and discovery_object.registry.has_widget(description.qualified_name)
        ):
            return
        discovery_object.handle_widget(description)



import logging
from AnyQt.QtCore import QThread
from ewokscore.task import TaskInputError
from ewokscore import TaskWithProgress


_logger = logging.getLogger(__name__)


class TaskExecutor:
    """Create and execute an Ewoks task"""

    def __init__(self, ewokstaskclass):
        self.__ewokstaskclass = ewokstaskclass
        self.__task = None

    def create_task(self, **kwargs):
        if not issubclass(self.__ewokstaskclass, TaskWithProgress):
            kwargs.pop("progress", None)
        self.__task = None
        try:
            self.__task = self.__ewokstaskclass(**kwargs)
        except TaskInputError:
            pass

    def execute_task(self):
        if not self.is_ready_to_execute:
            return
        try:
            self.__task.execute()
        except TaskInputError as e:
            _logger.warning(e)

    @property
    def is_ready_to_execute(self):
        if self.__task is None:
            return False
        return self.__task.is_ready_to_execute

    @property
    def succeeded(self):
        if self.__task is None:
            return False
        return self.__task.done and not self.__task.failed

    @property
    def output_variables(self):
        if self.__task is None:
            return dict()
        return self.__task.output_variables


class ThreadedTaskExecutor(QThread, TaskExecutor):
    """Create and execute an Ewoks task in a dedicated thread."""

    def run(self):
        self.execute_task()

    def stop(self, timeout=None, wait=False):
        self.blockSignals(True)
        if wait:
            if timeout:
                self.wait(timeout * 1000)
            else:
                self.wait()
        if self.isRunning():
            self.quit()



import os
import sys
import tempfile
from typing import Optional

from ewokscore import load_graph
from .owsconvert import ewoks_to_ows
from ..canvas.__main__ import main as launchcanvas


__all__ = ["execute_graph"]


def execute_graph(graph, load_options: Optional[dict] = None, **execute_options):
    if load_options is None:
        load_options = dict()
    ewoksgraph = load_graph(source=graph, **load_options)
    if ewoksgraph.is_cyclic:
        raise RuntimeError("Orange can only execute DAGs")
    if ewoksgraph.has_conditional_links:
        raise RuntimeError("Orange cannot handle conditional links")

    # We do not have a mapping between OWS and the runtime representation.
    # So map to a (temporary) persistent representation first.
    with tempfile.TemporaryDirectory() as tmpdirname:
        filename = os.path.join(tmpdirname, "ewokstaskgraph.ows")
        ewoks_to_ows(ewoksgraph, filename, **execute_options)
        argv = [sys.argv[0], filename]
        launchcanvas(argv=argv)



from AnyQt import QtWidgets
from AnyQt.QtCore import Qt
import numbers
from ewokscore.hashing import UniversalHashable


class ParameterForm(QtWidgets.QWidget):
    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent)
        self.initUI(**kwargs)
        self._fields = dict()

    def initUIParent(self):
        parent = self.parent()
        if parent is None:
            return
        layout = parent.layout()
        if layout is None:
            layout = QtWidgets.QVBoxLayout()
            parent.setLayout(layout)
        layout.addWidget(self)

    def initUI(self, margin=0, spacing=4):
        self.initUIParent()
        layout = QtWidgets.QFormLayout()
        layout.setContentsMargins(margin, margin, margin, margin)
        layout.setSpacing(spacing)
        policy = QtWidgets.QFormLayout.AllNonFixedFieldsGrow
        layout.setFieldGrowthPolicy(policy)
        # layout.setFormAlignment(Qt.AlignHCenter | Qt.AlignTop)
        layout.setLabelAlignment(Qt.AlignLeft)
        self.setLayout(layout)

    # def sizeHint(self):
    #    s = super().sizeHint()
    #    return QtCore.QSize(int(1.5 * s.width()), s.height())

    def addParameter(self, name, value=None, default="", changeCallback=None):
        label = name + ":"
        if value is None or value is UniversalHashable.MISSING_DATA:
            value = default
        if isinstance(value, str):
            field = QtWidgets.QLineEdit()
            field.setText(value)
            if changeCallback:
                field.textChanged.connect(changeCallback)
            else:
                field.setReadOnly(True)
        elif isinstance(value, numbers.Number):
            if isinstance(value, numbers.Integral):
                field = QtWidgets.QSpinBox()
                field.setRange(-(2**31), 2**31 - 1)
            else:
                field = QtWidgets.QDoubleSpinBox()
            field.setValue(value)
            if changeCallback:
                field.editingFinished.connect(changeCallback)
            else:
                field.setReadOnly(True)
        else:
            raise ValueError(
                "{} does not have a Qt widget".format(repr(type(value).__qualname__))
            )
        # Append
        policy = QtWidgets.QSizePolicy.Expanding
        field.setSizePolicy(policy, policy)
        self.layout().addRow(label, field)
        self._fields[name] = field

    def getParameter(self, name):
        field = self._fields[name]
        if isinstance(field, QtWidgets.QLineEdit):
            return field.text()
        else:
            return field.value()

    def setParameter(self, name, value):
        field = self._fields[name]
        if isinstance(field, QtWidgets.QLineEdit):
            field.setText(str(value))
        else:
            field.setValue(value)

    def enable(self, name):
        if name in self._fields:
            field = self._fields[name]
            field.setEnabled(True)

    def disable(self, name):
        if name in self._fields:
            field = self._fields[name]
            field.setEnabled(False)

    def getParameters(self):
        return {name: self.getParameter(name) for name in self._fields}



"""
Orange widget base classes to execute Ewoks tasks
"""
import inspect
import logging
from contextlib import contextmanager

from ..orange_version import ORANGE_VERSION

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    from oasys.widgets.widget import OWWidget
    from orangewidget.widget import WidgetMetaClass
    from orangewidget.settings import Setting

    OWBaseWidget = OWWidget
    summarize = None
    PartialSummary = None
    has_progress_bar = True
else:
    from orangewidget.widget import OWBaseWidget
    from orangewidget.settings import Setting
    from orangewidget.utils.signals import summarize
    from orangewidget.utils.signals import PartialSummary

    if ORANGE_VERSION == ORANGE_VERSION.latest_orange:
        from Orange.widgets.widget import OWWidget
        from Orange.widgets.widget import WidgetMetaClass

        has_progress_bar = True
    else:
        OWWidget = OWBaseWidget
        WidgetMetaClass = type(OWBaseWidget)
        has_progress_bar = False

from ewokscore.variable import Variable
from ewokscore.variable import value_from_transfer
from ewokscore.hashing import UniversalHashable
from ewokscore.hashing import MissingData
from .progress import QProgress
from .taskexecutor import TaskExecutor
from .taskexecutor import ThreadedTaskExecutor
from .taskexecutor_queue import TaskExecutorQueue
from . import owsignals


_logger = logging.getLogger(__name__)


__all__ = [
    "OWEwoksWidgetNoThread",
    "OWEwoksWidgetOneThread",
    "OWEwoksWidgetOneThreadPerRun",
    "OWEwoksWidgetWithTaskStack",
]


MISSING_DATA = UniversalHashable.MISSING_DATA
INVALIDATION_DATA = None  # This means we cannot use `None` as a value


if summarize is not None:

    @summarize.register(Variable)
    def summarize_variable(var: Variable):
        if var.value is MISSING_DATA:
            dtype = MISSING_DATA
        else:
            dtype = type(var.value).__name__
        desc = f"ewoks variable ({dtype})"
        return PartialSummary(desc, desc)

    @summarize.register(object)
    def summarize_object(value: object):
        return PartialSummary(str(type(value)), str(type(value)))


def prepare_OWEwoksWidgetclass(namespace, ewokstaskclass):
    """This needs to be called before signal and setting parsing"""
    namespace["ewokstaskclass"] = ewokstaskclass
    namespace["default_inputs"] = Setting(
        {name: INVALIDATION_DATA for name in ewokstaskclass.input_names()},
        schema_only=True,
    )
    namespace["varinfo"] = Setting(dict(), schema_only=True)
    owsignals.validate_inputs(namespace)
    owsignals.validate_outputs(namespace)


class _OWEwoksWidgetMetaClass(WidgetMetaClass):
    def __new__(metacls, name, bases, attrs, ewokstaskclass=None, **kw):
        if ewokstaskclass:
            prepare_OWEwoksWidgetclass(attrs, ewokstaskclass)
        return super().__new__(metacls, name, bases, attrs, **kw)


# insure compatibility between old orange widget and new
# orangewidget.widget.WidgetMetaClass. This was before split of the two
# projects. Parameter name "openclass" is undefined on the old version
ow_build_opts = {}
if "openclass" in inspect.signature(WidgetMetaClass).parameters:
    ow_build_opts["openclass"] = True


class OWEwoksBaseWidget(OWWidget, metaclass=_OWEwoksWidgetMetaClass, **ow_build_opts):
    """
    Base class to handle boiler plate code to interconnect ewoks and
    orange3
    """

    MISSING_DATA = MISSING_DATA

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dynamic_inputs = dict()
        self.__task_output_changed_callbacks = {self.task_output_changed}

    @classmethod
    def input_names(cls):
        return cls.ewokstaskclass.input_names()

    @classmethod
    def output_names(cls):
        return cls.ewokstaskclass.output_names()

    def _getTaskArguments(self):
        return {"inputs": self.task_inputs, "varinfo": self.varinfo}

    @staticmethod
    def _get_value(value):
        if isinstance(value, Variable):
            return value.value
        if isinstance(value, MissingData):
            # `Setting` seems to make a copy of MISSING_DATA
            return MISSING_DATA
        return value

    @property
    def defined_default_input_values(self) -> dict:
        # Warning: do not use default_inputs directly because it
        #          messes up MISSING_DATA
        return {
            name: value
            for name, value in self.default_inputs.items()
            if value is not INVALIDATION_DATA
        }

    @property
    def default_input_values(self) -> dict:
        values = {name: MISSING_DATA for name in self.input_names()}
        values.update(self.defined_default_input_values)
        return values

    @property
    def dynamic_input_values(self) -> dict:
        return {k: self._get_value(v) for k, v in self.__dynamic_inputs.items()}

    @property
    def task_inputs(self) -> dict:
        inputs = self.defined_default_input_values
        inputs.update(self.__dynamic_inputs)
        return inputs

    def receiveDynamicInputs(self, name, value):
        if value is INVALIDATION_DATA:
            self.__dynamic_inputs.pop(name, INVALIDATION_DATA)
        else:
            self.__dynamic_inputs[name] = value

    def _get_output_signal(self, ewoksname):
        if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
            for signal in self.outputs:
                if signal.name == ewoksname:
                    break
            else:
                signal = None
        else:
            signal = getattr(self.Outputs, ewoksname, None)
        if signal is None:
            raise RuntimeError(f"Output signal '{ewoksname}' does not exist")
        return signal

    def trigger_downstream(self):
        if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
            for ewoksname, var in self.task_outputs.items():
                ewoks_to_orange = owsignals.get_ewoks_to_orange_mapping(
                    type(self), "outputs"
                )
                orangename = ewoks_to_orange.get(ewoksname, ewoksname)
                if var.value is MISSING_DATA:
                    self.send(orangename, INVALIDATION_DATA)  # or channel.invalidate?
                else:
                    self.send(orangename, var)
        else:
            for ewoksname, var in self.task_outputs.items():
                channel = self._get_output_signal(ewoksname)
                if var.value is MISSING_DATA:
                    channel.send(INVALIDATION_DATA)  # or channel.invalidate?
                else:
                    channel.send(var)
        for cb in self.__task_output_changed_callbacks:
            cb()

    @property
    def task_output_changed_callbacks(self) -> set:
        return self.__task_output_changed_callbacks

    def task_output_changed(self):
        pass

    def clear_downstream(self):
        if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
            for name in self.task_outputs:
                self.send(name, INVALIDATION_DATA)  # or channel.invalidate?
        else:
            for name in self.task_outputs:
                channel = self._get_output_signal(name)
                channel.send(INVALIDATION_DATA)  # or channel.invalidate?

    def defaultInputsHaveChanged(self):
        """Needs to be called when default inputs have changed"""
        self.executeEwoksTask()

    def handleNewSignals(self):
        """Invoked by the workflow signal propagation manager after all
        signals handlers have been called.
        """
        self.executeEwoksTask()

    def executeEwoksTask(self):
        raise NotImplementedError("Base class")

    @property
    def task_outputs(self):
        raise NotImplementedError("Base class")

    @property
    def task_output_values(self):
        return {name: var.value for name, var in self.task_outputs.items()}

    @property
    def task_input_values(self):
        return {
            name: value_from_transfer(var, varinfo=self.varinfo)
            for name, var in self.task_inputs.items()
        }


def is_orange_widget_class(widget_class):
    return issubclass(widget_class, OWBaseWidget)


def is_ewoks_widget_class(widget_class):
    return issubclass(widget_class, OWEwoksBaseWidget)


def is_native_widget_class(widget_class):
    return is_orange_widget_class(widget_class) and not is_ewoks_widget_class(
        widget_class
    )


def is_orange_widget(widget):
    return isinstance(widget, OWBaseWidget)


def is_ewoks_widget(widget):
    return isinstance(widget, OWEwoksBaseWidget)


def is_native_widget(widget_class):
    return is_orange_widget(widget_class) and not is_ewoks_widget(widget_class)


class OWEwoksWidgetNoThread(OWEwoksBaseWidget, **ow_build_opts):
    """Widget which will executeEwoksTask the ewokscore.Task directly"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__taskExecutor = TaskExecutor(self.ewokstaskclass)

    def executeEwoksTask(self):
        self.__taskExecutor.create_task(**self._getTaskArguments())
        try:
            self.__taskExecutor.execute_task()
        except Exception:
            self.clear_downstream()
            raise
        if self.__taskExecutor.succeeded:
            self.trigger_downstream()
        else:
            self.clear_downstream()

    @property
    def task_outputs(self):
        return self.__taskExecutor.output_variables


class _OWEwoksThreadedBaseWidget(OWEwoksBaseWidget, **ow_build_opts):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__taskProgress = QProgress()
        if has_progress_bar:
            self.__taskProgress.sigProgressChanged.connect(self.progressBarSet)

    def onDeleteWidget(self):
        if has_progress_bar:
            self.__taskProgress.sigProgressChanged.disconnect(self.progressBarSet)
        self._cleanupTaskExecutor()
        super().onDeleteWidget()

    def _cleanupTaskExecutor(self):
        raise NotImplementedError("Base class")

    @contextmanager
    def _ewoksTaskStartContext(self):
        try:
            self.__ewoksTaskInit()
            yield
        except Exception:
            self.__ewoksTaskFinished()
            raise

    @contextmanager
    def _ewoksTaskFinishedContext(self):
        try:
            yield
        finally:
            self.__ewoksTaskFinished()

    def __ewoksTaskInit(self):
        self.progressBarInit()

    def __ewoksTaskFinished(self):
        self.progressBarFinished()

    def _getTaskArguments(self):
        adict = super()._getTaskArguments()
        adict["progress"] = self.__taskProgress
        return adict


class OWEwoksWidgetOneThread(_OWEwoksThreadedBaseWidget, **ow_build_opts):
    """
    All the processing is done on one thread.
    If a processing is requested when the thread is already running then
    it is refused.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__taskExecutor = ThreadedTaskExecutor(ewokstaskclass=self.ewokstaskclass)
        self.__taskExecutor.finished.connect(self._ewoksTaskFinishedCallback)

    def executeEwoksTask(self):
        if self.__taskExecutor.isRunning():
            _logger.error("A processing is already on going")
            return
        else:
            self.__taskExecutor.create_task(**self._getTaskArguments())
            if self.__taskExecutor.is_ready_to_execute:
                with self._ewoksTaskStartContext():
                    self.__taskExecutor.start()

    @property
    def task_outputs(self):
        return self.__taskExecutor.output_variables

    def _ewoksTaskFinishedCallback(self):
        with self._ewoksTaskFinishedContext():
            if self.__taskExecutor.succeeded:
                self.trigger_downstream()
            else:
                self.clear_downstream()

    def _cleanupTaskExecutor(self):
        self.__taskExecutor.finished.disconnect(self._ewoksTaskFinishedCallback)
        self.__taskExecutor.stop()
        self.__taskExecutor = None


class OWEwoksWidgetOneThreadPerRun(_OWEwoksThreadedBaseWidget, **ow_build_opts):
    """
    Each time a task processing is requested this will create a new thread
    to do the processing.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__taskExecutors = list()
        self.__last_output_variables = dict()

    def executeEwoksTask(self):
        taskExecutor = ThreadedTaskExecutor(ewokstaskclass=self.ewokstaskclass)
        taskExecutor.create_task(**self._getTaskArguments())
        if not taskExecutor.is_ready_to_execute:
            return
        with self.__addTaskExecutor(taskExecutor):
            with self._ewoksTaskStartContext():
                taskExecutor.start()

    @contextmanager
    def __addTaskExecutor(self, taskExecutor):
        self.__disconnectAllTaskExecutors()
        taskExecutor.finished.connect(self._ewoksTaskFinishedCallback)
        self.__taskExecutors.append(taskExecutor)
        try:
            yield
        except Exception:
            taskExecutor.finished.disconnect(self._ewoksTaskFinishedCallback)
            self.__taskExecutors.remove(taskExecutor)
            raise

    def __disconnectAllTaskExecutors(self):
        for taskExecutor in self.__taskExecutors:
            try:
                taskExecutor.finished.disconnect(self._ewoksTaskFinishedCallback)
            except KeyError:
                pass

    def _ewoksTaskFinishedCallback(self):
        with self._ewoksTaskFinishedContext():
            try:
                taskExecutor = self.sender()
                self.__last_output_variables = taskExecutor.output_variables
                if taskExecutor.succeeded:
                    self.trigger_downstream()
                else:
                    self.clear_downstream()
            finally:
                if taskExecutor in self.__taskExecutors:
                    self.__taskExecutors.remove(taskExecutor)

    def _cleanupTaskExecutor(self):
        self.__disconnectAllTaskExecutors()
        for taskExecutor in self.__taskExecutors:
            taskExecutor.quit()
        self.__taskExecutors.clear()

    @property
    def task_outputs(self):
        return self.__last_output_variables


class OWEwoksWidgetWithTaskStack(_OWEwoksThreadedBaseWidget, **ow_build_opts):
    """
    Each time a task processing is requested add it to the FIFO stack.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__taskExecutorQueue = TaskExecutorQueue(ewokstaskclass=self.ewokstaskclass)
        self.__last_output_variables = dict()

    def executeEwoksTask(self):
        with self._ewoksTaskStartContext():
            self.__taskExecutorQueue.add(
                _callbacks=(self._ewoksTaskFinishedCallback,),
                **self._getTaskArguments(),
            )

    @property
    def task_outputs(self):
        return self.__last_output_variables

    def _cleanupTaskExecutor(self):
        self.__taskExecutorQueue.stop()
        self.__taskExecutorQueue = None

    def _ewoksTaskFinishedCallback(self):
        with self._ewoksTaskFinishedContext():
            taskExecutor = self.sender()
            self.__last_output_variables = taskExecutor.output_variables
            if taskExecutor.succeeded:
                self.trigger_downstream()
            else:
                self.clear_downstream()



from ewoksorange.setuptools import setup

if __name__ == "__main__":
    setup(
        __file__,
        name="ewoks-example-2-addon",
        version="0.0.1a",
        description="Add-on containing Ewoks example widgets",
    )



import pytest
from ewoksorange.bindings import OWWIDGET_TASKS_GENERATOR
from ewokscore.task import TaskInputError
from ewokscore.inittask import instantiate_task


@pytest.mark.parametrize(
    "widget_qualname",
    [
        "orangecontrib.ewoks_example_supercategory.ewoks_example_subcategory.adder1.Adder1",
        "orangecontrib.ewoks_example_supercategory.ewoks_example_subcategory.adder2.Adder2",
        "orangecontrib.evaluate.ewoks_example_submodule.adder1.Adder1",
        "orangecontrib.evaluate.ewoks_example_submodule.adder2.Adder2",
        "orangecontrib.ewoks_example_category.adder1.Adder1",
        "orangecontrib.ewoks_example_category.adder2.Adder2",
    ],
)
def test_adder_missing_inputs(widget_qualname, register_ewoks_example_1_addon):
    node_attrs = {
        "task_type": "generated",
        "task_identifier": widget_qualname,
        "task_generator": OWWIDGET_TASKS_GENERATOR,
    }
    with pytest.raises(TaskInputError):
        instantiate_task(node_attrs)


@pytest.mark.parametrize(
    "widget_qualname",
    [
        "orangecontrib.ewoks_example_supercategory.ewoks_example_subcategory.adder1.Adder1",
        "orangecontrib.ewoks_example_supercategory.ewoks_example_subcategory.adder2.Adder2",
        "orangecontrib.evaluate.ewoks_example_submodule.adder1.Adder1",
        "orangecontrib.evaluate.ewoks_example_submodule.adder2.Adder2",
        "orangecontrib.ewoks_example_category.adder1.Adder1",
        "orangecontrib.ewoks_example_category.adder2.Adder2",
    ],
)
def test_adder_all_inputs(widget_qualname, register_ewoks_example_1_addon):
    node_attrs = {
        "task_type": "generated",
        "task_identifier": widget_qualname,
        "task_generator": OWWIDGET_TASKS_GENERATOR,
    }
    task = instantiate_task(node_attrs, inputs={"a": 1, "b": 2})
    task.execute()
    assert task.output_values == {"result": 3}



import os
import tempfile
import pytest

from ewokscore.tests.examples.graphs import graph_names
from ewokscore.tests.examples.graphs import get_graph
from ewokscore.tests.utils.results import assert_execute_graph_all_tasks
from ewokscore import load_graph
from ewoksorange.bindings import ewoks_to_ows


@pytest.mark.parametrize("graph_name", graph_names())
def test_execute_graph(graph_name, tmpdir, ewoks_orange_canvas):
    """Test graph execution like the Orange canvas would do it"""
    graph, expected = get_graph(graph_name)
    ewoksgraph = load_graph(graph)
    varinfo = {"root_uri": str(tmpdir)}
    no_explicit_datamapping = any(
        not link_attrs.get("data_mapping")
        for link_attrs in ewoksgraph.graph.edges.values()
    )
    if (
        ewoksgraph.is_cyclic
        or ewoksgraph.has_conditional_links
        or no_explicit_datamapping
    ):
        pytest.skip("graph not supported by orange")

    with tempfile.TemporaryDirectory() as tmpdirname:
        filename = os.path.join(tmpdirname, graph_name + ".ows")
        ewoks_to_ows(ewoksgraph, filename, varinfo=varinfo, error_on_duplicates=False)
        ewoks_orange_canvas.load_ows(filename)
    ewoks_orange_canvas.wait_widgets(timeout=10)

    assert_execute_graph_all_tasks(ewoksgraph, expected, varinfo=varinfo)



from AnyQt.QtWidgets import QPushButton
from ewoksorange.orange_version import ORANGE_VERSION

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    from oasys.widgets import gui
elif ORANGE_VERSION == ORANGE_VERSION.latest_orange:
    from Orange.widgets import gui
else:
    from orangewidget import gui
from ewoksorange.bindings import OWEwoksWidgetNoThread
from ewoksorange.gui.parameterform import ParameterForm
from ewoksorange.tests.listoperations import GenerateList


class ListGenerator(OWEwoksWidgetNoThread, ewokstaskclass=GenerateList):
    name = "List generator"
    description = "Generate a random list with X elements"
    icon = "icons/mywidget.svg"
    want_main_area = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        box = gui.widgetBox(self.controlArea, "Default Inputs")
        self._default_inputs_form = ParameterForm(parent=box)
        for name, value in self.default_input_values.items():
            self._default_inputs_form.addParameter(
                name,
                value=value,
                default=0,
                changeCallback=self.defaultInputsHaveChanged,
            )

        box = gui.widgetBox(self.controlArea, "Dynamic Inputs")
        self._dynamic_input_form = ParameterForm(parent=box)
        for name in self.input_names():
            self._dynamic_input_form.addParameter(name)

        box = gui.widgetBox(self.controlArea, "Outputs")
        self._output_form = ParameterForm(parent=box)
        for name in self.output_names():
            self._output_form.addParameter(name)

        box = gui.widgetBox(self.controlArea, "Commands")
        layout = box.layout()
        self._validateButton = QPushButton("generate", self)
        layout.addWidget(self._validateButton)

        # connect signal / slot
        self._validateButton.released.connect(self.handleNewSignals)

    def defaultInputsHaveChanged(self):
        self.default_inputs.update(self._default_inputs_form.getParameters())
        super().defaultInputsHaveChanged()
        for name, value in self.task_output_values.items():
            self._output_form.setParameter(name, value)

    def handleNewSignals(self):
        for name, value in self.dynamic_input_values.items():
            self._dynamic_input_form.setParameter(name, value)
            self._default_inputs_form.disable(name)
        super().handleNewSignals()
        for name, value in self.task_output_values.items():
            self._output_form.setParameter(name, value)



import pytest
from ewoksorange.bindings import OWWIDGET_TASKS_GENERATOR
from ewokscore.inittask import instantiate_task


@pytest.mark.parametrize(
    "widget_qualname",
    [
        "orangecontrib.list_operations.sumlist_one_thread.SumListOneThread",
        "orangecontrib.list_operations.sumlist_several_thread.SumListSeveralThread",
        "orangecontrib.list_operations.sumlist_stack.SumListWithTaskStack",
    ],
)
def test_sumlist(widget_qualname, register_ewoks_example_2_addon):
    node_attrs = {
        "task_type": "generated",
        "task_identifier": widget_qualname,
        "task_generator": OWWIDGET_TASKS_GENERATOR,
    }
    task = instantiate_task(node_attrs, inputs={"list": [1, 2, 3]})
    task.execute()
    assert task.output_values == {"sum": 6}


def test_listgenerator(register_ewoks_example_2_addon):
    node_attrs = {
        "task_type": "generated",
        "task_identifier": "orangecontrib.list_operations.listgenerator.ListGenerator",
        "task_generator": OWWIDGET_TASKS_GENERATOR,
    }
    task = instantiate_task(node_attrs, inputs={"length": [7]})
    task.execute()
    assert len(task.output_values["list"]) == 7


def test_printsum(register_ewoks_example_2_addon):
    node_attrs = {
        "task_type": "generated",
        "task_identifier": "orangecontrib.list_operations.print_sum.PrintSumOW",
        "task_generator": OWWIDGET_TASKS_GENERATOR,
    }
    task = instantiate_task(node_attrs, inputs={"sum": 99})
    task.execute()
    assert task.succeeded
    assert task.output_values == {}



"""Main entry point

    python -m ewoksorange.canvas --with_example

Which is equivalent to

    python -m Orange.canvas

but it registers the example Addon before launching.
"""

import sys
import logging
from contextlib import contextmanager

from ..orange_version import ORANGE_VERSION

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    from oasys.canvas.__main__ import main as orange_main
elif ORANGE_VERSION == ORANGE_VERSION.latest_orange:
    from Orange.canvas.__main__ import main as orange_main
else:
    from orangecanvas.main import Main as OrangeCanvasMain

    class Main(OrangeCanvasMain):
        DefaultConfig = "ewoksorange.canvas.config.Config"

    def orange_main(argv):
        main_instance = Main()
        return main_instance.run(argv)


if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    import argparse

    def arg_parser():
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "--no-discovery",
            action="store_true",
            help="Don't run widget discovery (use full cache instead)",
        )
        parser.add_argument(
            "--force-discovery",
            action="store_true",
            help="Force full widget discovery (invalidate cache)",
        )
        parser.add_argument(
            "--no-welcome", action="store_true", help="Don't show welcome dialog."
        )
        parser.add_argument(
            "--no-splash", action="store_true", help="Don't show splash screen."
        )
        parser.add_argument(
            "--stylesheet",
            help="Application level CSS style sheet to use",
            type=str,
            default=None,
        )
        parser.add_argument(
            "--config",
            help="Configuration namespace",
            type=str,
            default=None,
        )
        parser.add_argument(
            "-l",
            "--log-level",
            help="Logging level (0, 1, 2, 3, 4)",
            type=int,
            default=1,
        )
        return parser

else:
    from orangecanvas.main import arg_parser

from ewoksorange.registration import register_addon_package
from ewoksorange import ewoks_addon


@contextmanager
def temporary_log_handlers(log_level):
    logger = logging.getLogger("ewoksorange")
    logger.setLevel(log_level)
    if logger.hasHandlers():
        yield
    else:
        stdouthandler = logging.StreamHandler(sys.stdout)
        logger.addHandler(stdouthandler)
        yield
        logger.removeHandler(stdouthandler)


def main(argv=None):
    parser = arg_parser()
    parser.add_argument(
        "--with_example",
        action="store_true",
        help="Register example add-on's from ewoksorange.",
    )

    if argv is None:
        argv = sys.argv
    options, _ = parser.parse_known_args(argv[1:])

    if "--with_example" in argv:
        argv.pop(argv.index("--with_example"))

    if "--force-discovery" not in argv:
        argv.append("--force-discovery")

    with temporary_log_handlers(options.log_level):
        register_addon_package(ewoks_addon)

        if options.with_example:
            from ewoksorange.tests.examples import ewoks_example_1_addon
            from ewoksorange.tests.examples import ewoks_example_2_addon

            register_addon_package(ewoks_example_1_addon)
            register_addon_package(ewoks_example_2_addon)

    orange_main(argv)


if __name__ == "__main__":
    sys.exit(main())



"""Copy parts of Orange.canvas.config to be used when Orange3 is not installed.
"""

import pkg_resources
from orangewidget.workflow import config as orangeconfig
from orangewidget.workflow.config import WIDGETS_ENTRY


class Config(orangeconfig.Config):
    @staticmethod
    def widgets_entry_points():
        """
        Return an `EntryPoint` iterator for all 'orange.widget' entry
        points.
        """
        # Ensure the 'this' distribution's ep is the first. iter_entry_points
        # yields them in unspecified order.
        all_eps = pkg_resources.iter_entry_points(WIDGETS_ENTRY)
        return iter(all_eps)

    @staticmethod
    def addon_entry_points():
        return Config.widgets_entry_points()


def widgets_entry_points():
    return Config.widgets_entry_points()



from ewoksorange.orange_version import ORANGE_VERSION

if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
    from oasys.widgets import gui
elif ORANGE_VERSION == ORANGE_VERSION.latest_orange:
    from Orange.widgets import gui
    from Orange.widgets.widget import Input, Output
else:
    from orangewidget import gui
    from orangewidget.widget import Input, Output

from ewoksorange.bindings import OWEwoksWidgetNoThread
from ewoksorange.gui.parameterform import ParameterForm
from ewokscore.tests.examples.tasks.sumtask import SumTask


__all__ = ["Adder1"]


class Adder1(OWEwoksWidgetNoThread, ewokstaskclass=SumTask):
    name = "Adder1"
    description = "Adds two numbers"
    icon = "icons/mywidget.svg"
    want_main_area = False

    if ORANGE_VERSION == ORANGE_VERSION.oasys_fork:
        inputs = [("A", object, ""), ("B", object, "")]
        outputs = [{"name": "A + B", "id": "A + B", "type": object}]
        inputs_orange_to_ewoks = {"A": "a", "B": "b"}
        outputs_orange_to_ewoks = {"A + B": "result"}
    else:

        class Inputs:
            a = Input("A", object)
            b = Input("B", object)

        class Outputs:
            result = Output("A + B", object)

    def __init__(self):
        super().__init__()

        box = gui.widgetBox(self.controlArea, "Default Inputs")
        self._default_inputs_form = ParameterForm(parent=box)
        for name, value in self.default_input_values.items():
            self._default_inputs_form.addParameter(
                name,
                value=value,
                default=0,
                changeCallback=self.defaultInputsHaveChanged,
            )

        box = gui.widgetBox(self.controlArea, "Dynamic Inputs")
        self._dynamic_input_form = ParameterForm(parent=box)
        for name in self.input_names():
            self._dynamic_input_form.addParameter(name)

        box = gui.widgetBox(self.controlArea, "Outputs")
        self._output_form = ParameterForm(parent=box)
        for name in self.output_names():
            self._output_form.addParameter(name)

        self.handleNewSignals()

    def defaultInputsHaveChanged(self):
        self.default_inputs.update(self._default_inputs_form.getParameters())
        super().defaultInputsHaveChanged()
        for name, value in self.task_output_values.items():
            self._output_form.setParameter(name, value)

    def handleNewSignals(self):
        for name, value in self.dynamic_input_values.items():
            self._dynamic_input_form.setParameter(name, value)
            self._default_inputs_form.disable(name)
        super().handleNewSignals()
        for name, value in self.task_output_values.items():
            self._output_form.setParameter(name, value)
